grammar FooLang;

@header {
import cz.upol.foolang.ast.*;
import cz.upol.foolang.types.*;
import java.util.*;
}

@parser::members {
	
	/** trida sestavujici AST */
	ASTBuilder ast = new ASTBuilder();
	
	/** aktualni balik souboru, definovany 'package' */
	List<String> classPackage = new ArrayList<>();

	/** nazev tridy */
	String className;

	/** jednotlive typy importovane do dane tridy */
	List<GeneralType> imports = new ArrayList<>();

	/** 
         * pomocna procedura provadejici konverzi escapovanych znaku
         * na jejich skutecne hodnoty.
         */  
	public static String unescapeString(String s) {
		StringBuilder result = new StringBuilder();
		for (int i = 0; i < s.length(); i++) {
			char c1 = s.charAt(i);
			if ((c1 == '\\') && (i + 1 < s.length())) {
				char c2 = s.charAt(i + 1);
				switch (c2) {
					case 'n': result.append('\n'); i++; break;
					case 't': result.append('\t'); i++; break;
					case 'r': result.append('\r'); i++; break;
					case '\\': result.append('\\'); i++; break;
					default: result.append(c1);
				}
			} else result.append(c1);
		}
		return result.toString();
	}

	/**
         * Pro zvoleny typ vraci prislusny balik
         */
	public List<String> getPackage(String name) {
		if (name.equals(className)) return classPackage;
		
		// hleda prvni vhodny import
		for (GeneralType anImport : imports)
			if (anImport.getTypeName().equals(name)) 
				return anImport.getTypePackage();
		
		// automaticky importuje soubory z java.lang
		try {
			// pokud to neskonci vyjimkou, trida existuje v java.lang
			Class.forName("java.lang." + name);
			return Arrays.asList("java", "lang");
		} catch (ClassNotFoundException e) {
			// ignore this
		}
		throw new RuntimeException("Unknown location of: " + name);
	}
}

// zdrojovy kod, skladajici se z balicku, importu a deklarace tridy
source returns [Node node]
	: (PACKAGE (pkg=IDENTIFIER '.' { classPackage.add($pkg.text); })*
	           pkg=IDENTIFIER { classPackage.add($pkg.text); } ';') ?
	  (IMPORT type { imports.add((GeneralType) $type.name); } ';')*
          classDecl { $node = $classDecl.node; }
        ; 

// deklarace tridy: class Foo { atributy* metody* }
classDecl returns [Node node]
	: { List<Node> ll = new ArrayList<>(); }
	   CLASS name=IDENTIFIER { className = $name.text; }
	 '{' 
              (fieldDecl  { ll.add($fieldDecl.node); })* 
              (methodDecl { ll.add($methodDecl.node); })* 
	 '}'
	  { $node = ast.createClass($name.text, ll); }
	;

// deklarace metody: [modifikatory] nazev
fieldDecl returns [Node node]
	: modifiers type name=IDENTIFIER ';'
		{ $node = ast.createField($name.text, $modifiers.value, $type.name); }
	;


// deklarace metody: [modifikatory] nazev(argument-1, argument-2, ...) { telo metody }
methodDecl returns [Node node]
	: modifiers type name=IDENTIFIER arguments block
		{ $node = ast.createMethod($name.text, $modifiers.value, $type.name, $arguments.arg, $block.node); }
	;



// predstavuje blok prikazu
block returns [Node node]
	: 	{ List<Node> ll = new ArrayList<>(); }
		'{' (statement { ll.add($statement.node); })* '}'
		{ $node = ast.createBlock(ll); }
	;

// prikaz, v soucasne dobe je podporovany jen prikaz deklarace promenne a prikaz 'return'
statement returns [Node node]
	: varDecl ';'		{ $node = $varDecl.node; }
	| RETURN expr ';'	{ $node = ast.createReturn($expr.node); }
	;

// deklarace promenne; neni mozna inicializace
varDecl returns [Node node]
	: type name=IDENTIFIER	{ $node = ast.createVarDecl($name.text, $type.name); }
	;

// seznam argumentu
arguments returns [List<Argument> arg]
	: 	{ List<Argument> ll = new ArrayList<>(); }
		'(' (argument     { ll.add($argument.arg); })? 
                    (',' argument { ll.add($argument.arg); })* ')'
		{ $arg = ll; }
	;

// argument metody, kazdy argument ma svuj typ a nazev
argument returns [Argument arg]
	: type name=IDENTIFIER	{ $arg = ast.createArgument($type.name, $name.text); }
	;

// modifikatory metod a atributu, rozlisujeme dva typy:
// (1) membership (jestli se metoda atribut vztahuje k objektu nebo tride)
// (2) scope (jinymi slovy viditelnost -- public, protected, private)
modifiers returns [Set<Modifier> value]
	: { Set<Modifier> ss = new HashSet<>(); }
	  (sd=scopeDecl     { ss.add($sd.value); })
	  (md=membershipDecl { ss.add($md.value); })?
	  { $value = ss; }

	| { Set<Modifier> ss = new HashSet<>(); }
	   (md=membershipDecl { ss.add($md.value); })?
	   (sd=scopeDecl      { ss.add($sd.value); })
	  { $value = ss; }
	;

scopeDecl returns [Modifier value]
	: PRIVATE	{ $value = Modifier.PRIVATE; }
	| PUBLIC	{ $value = Modifier.PUBLIC; } 
	| PROTECTED	{ $value = Modifier.PROTECTED; }	
	;

membershipDecl returns [Modifier value]
	: STATIC	{ $value = Modifier.STATIC; } 
	;

// typ: primitivni nebo obecny 
// (pokud neni uvedeny nazev typu vcetne balicku,
// uvazuje se aktualni balicek, pripodne importovane typy)
type returns [Type name]
	:
	| T_INT		{ $name = PrimitiveType.INT; }
	| T_BOOLEAN	{ $name = PrimitiveType.BOOLEAN; }
	| T_DOUBLE	{ $name = PrimitiveType.DOUBLE; }
	| T_CHAR	{ $name = PrimitiveType.CHAR; }
	| T_VOID	{ $name = PrimitiveType.VOID; }
	| { List<String> ll = new ArrayList<>(); }
	  (pkg=IDENTIFIER '.' { ll.add($pkg.text); })*
	  typeName=IDENTIFIER
	  { $name = ast.createType(ll.size() > 0 ? ll : getPackage($typeName.text), $typeName.text); }
	;

// vyraz
expr returns [Node node]
	: atom		      { $node = $atom.node; }
	| '(' expr ')'        { $node = $expr.node; }
	| e1=expr '=' e2=expr { $node = ast.createBinaryOperator(Operator.ASSIGN, $e1.node, $e2.node); }
	| e1=expr '+' e2=expr { $node = ast.createBinaryOperator(Operator.ADD, $e1.node, $e2.node); }
	| e1=expr '-' e2=expr { $node = ast.createBinaryOperator(Operator.SUB, $e1.node, $e2.node); }
	| e1=expr '*' e2=expr { $node = ast.createBinaryOperator(Operator.MUL, $e1.node, $e2.node); }
	| e1=expr '/' e2=expr { $node = ast.createBinaryOperator(Operator.DIV, $e1.node, $e2.node); }
	;

// atomicke vyrazy, tj. cisla, retezce, atp. + identifikatory promennych, atributu, atp. 
atom returns [Node node]
        : INT           { $node = ast.createConstant(Integer.parseInt($INT.text)); }
        | FLOAT         { $node = ast.createConstant(Double.parseDouble($FLOAT.text)); }
        | IDENTIFIER    { $node = ast.createVariable($IDENTIFIER.text); }
        | TRUE          { $node = ast.createConstant(true); }
        | FALSE         { $node = ast.createConstant(false); }
        | STRING        { $node = ast.createConstant(unescapeString($STRING.text.substring(1, $STRING.text.length() - 1))); }
        ;


// tokeny pro klicova slova a literaly
CLASS   : 'class';
PRIVATE : 'private';
PUBLIC  : 'public';
PROTECTED : 'protected';
STATIC  : 'static';
RETURN	: 'return';
PACKAGE : 'package';
IMPORT	: 'import';

T_INT	: 'int';
T_BOOLEAN: 'boolean';
T_DOUBLE: 'double';
T_CHAR  : 'char';
T_VOID	: 'void';

TRUE    : 'true';
FALSE   : 'false';
INT     : '-'?[0-9]+ ;
FLOAT   : '-'?[0-9]* '.' [0-9]+ ;
IDENTIFIER: IDENTIFIER_HEAD IDENTIFIER_REST*;
STRING: '"' ~('\n'|'"')* '"' ;
COMMENT : '//' ~('\r' | '\n')* -> skip;
WS      : [ \r\t\n]+ -> skip ;

// fragmenty zjednosujici tvorbu tokenu
fragment IDENTIFIER_HEAD
        :   'a'..'z' | 'A'..'Z' | '_'
        ;
    
fragment IDENTIFIER_REST
        : IDENTIFIER_HEAD
        | '0'..'9'
        ;


